﻿
using System;
using System.Threading;
namespace BeeDevelopment.PlayJoy {

	/// <summary>
	/// Provides methods for working with PlayStation controllers.
	/// </summary>
	public class Controller {

		/// <summary>
		/// Gets the base address of the parallel port that the controller is connected to.
		/// </summary>
		public short Address { get; private set; }

		/// <summary>
		/// Gets whether a device is connected.
		/// </summary>
		public bool Connected { get; private set; }

		/// <summary>
		/// Gets whether the device is in analogue mode.
		/// </summary>
		public bool Analog { get; private set; }

		/// <summary>
		/// Gets the position of the left joystick.
		/// </summary>
		public JoystickPosition LeftJoystick { get; private set; }

		/// <summary>
		/// Gets the position of the right joystick.
		/// </summary>
		public JoystickPosition RightJoystick { get; private set; }

		/// <summary>
		/// Gets the state of the currently pressed buttons.
		/// </summary>
		public Buttons ButtonState { get; private set; }

		/// <summary>
		/// Enumeration defining the bits used to 
		/// </summary>
		[Flags()] public enum Buttons {
			/// <summary>
			/// No buttons.
			/// </summary>
			None = 0,
			/// <summary>
			/// The Select button.
			/// </summary>
			Select = 1 << 0,
			/// <summary>
			/// The L3 shoulder button.
			/// </summary>
			L3 = 1 << 1,
			/// <summary>
			/// The R3 shoulder button.
			/// </summary>
			R3 = 1 << 2,
			/// <summary>
			/// The Start button.
			/// </summary>
			Start = 1 << 3,
			/// <summary>
			/// The Up directional pad button.
			/// </summary>
			Up = 1 << 4,
			/// <summary>
			/// The Right directional pad button.
			/// </summary>
			Right = 1 << 5,
			/// <summary>
			/// The Down directional pad button.
			/// </summary>
			Down = 1 << 6,
			/// <summary>
			/// The Left directional pad button.
			/// </summary>
			Left = 1 << 7,
			/// <summary>
			/// The L2 shoulder button.
			/// </summary>
			L2 = 1 << 8,
			/// <summary>
			/// The R2 shoulder button.
			/// </summary>
			R2 = 1 << 9,
			/// <summary>
			/// The L1 shoulder button.
			/// </summary>
			L1 = 1 << 10,
			/// <summary>
			/// The R1 shoulder button.
			/// </summary>
			R1 = 1 << 11,
			/// <summary>
			/// The Triangle symbol button.
			/// </summary>
			Triangle = 1 << 12,
			/// <summary>
			/// The Circle symbol button.
			/// </summary>
			Circle = 1 << 13,
			/// <summary>
			/// The Cross symbol button.
			/// </summary>
			Cross = 1 << 14,
			/// <summary>
			/// The Square symbol button.
			/// </summary>
			Square = 1 << 15,
			/// <summary>
			/// Bitmask for all buttons.
			/// </summary>
			/// <summary>
			/// Bitmask for all buttons.
			/// </summary>
			All = Select | Start | L3 | R3 | Up | Down | Left | Right | L2 | R2 | L1 | R1 | Triangle | Circle | Cross | Square,
		}

		/// <summary>
		/// Creates an instance of <see cref="Controller"/>.
		/// </summary>
		/// <param name="address">The base address of the parallel port that the controller is connected to.</param>
		public Controller(int address) {
			this.Address = (short)address;
			this.Update();
		}

		/// <summary>
		/// Gets a packet from the connected device.
		/// </summary>
		/// <param name="command">The command to send.</param>
		/// <param name="data">An optional array of data to send to the device.</param>
		/// <returns>The received packet.</returns>
		private byte[] GetPacket(byte command, byte[] data) {
			byte[] Packet = new byte[64];
			if (data != null) {
				Array.Copy(data, Packet, data.Length);
			}
			int Received = Interop.SendPacket(this.Address, ref command, Packet, Packet.Length);
			Array.Resize<byte>(ref Packet, Received);
			return Packet;			
		}

		/// <summary>
		/// Gets a packet from the connected device.
		/// </summary>
		/// <param name="command">The command to send.</param>
		/// <returns>The received packet.</returns>
		private byte[] GetPacket(byte command) {
			return this.GetPacket(command, null);
		}

		private bool Select {
			set { Interop.SetSelect(this.Address, value); }
		}

		/// <summary>
		/// Update the state of the controller.
		/// </summary>
		public void Update() {
			this.Select = false;
			byte[] Data = this.GetPacket(0x42);
			this.Select = true;
			this.Connected = Data.Length >= 2;
			
			if (!this.Connected) {
				this.LeftJoystick = new JoystickPosition();
				this.RightJoystick = new JoystickPosition();
				this.ButtonState = Buttons.None;
			} else {
				this.ButtonState = (Buttons)(~(Data[0] + Data[1] * 256)) & Buttons.All;
				this.Analog = Data.Length >= 6;
				if (this.Analog) {
					this.RightJoystick = new JoystickPosition(Data[2], Data[3]);
					this.LeftJoystick = new JoystickPosition(Data[4], Data[5]);
				}
			}
		}

	}
}
